// package JavaCode;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.Collections;
import java.util.Iterator;
import java.util.PriorityQueue;
import java.util.StringTokenizer;
import java.util.Vector;

/*
SLIDER: Local Short Reads DNA Sequence Alignment
Beta Release; Ver 0.1, 22th Mar. 2008
By: Nawar Malhis
Canada's Michael Smith Genome Sciences Centre
*/

public class OligoFile {
	
	static public final int iDfns = 0;
	static public final int iDfn  = 0;
	
	static public final int F = 0;
	static public final int S = 1;
	static public final int W = 0;
	static public final int R = 1;
	static public boolean display   = false;
	// count[s0/s1/s/s&][all/crisp/LQ/UM/UMCrisp/UMLQ]
	static public final int S0      = 0;
	static public final int S1      = 1;
	static public final int SAll    = 2;
	static public final int SAnd    = 3;
	static public final int SOr     = 4;
	static public final int ALL     = 0;
	static public final int Crisp   = 1;
	static public final int LQ      = 2;
	static public final int UM      = 3;
	static public final int UMCrisp = 4;
	static public final int UMLQ    = 5;

	// static DataInputStream strInPrb = null;
	public boolean prbLoaded = false;
	public String  vName;           // vertual file name
	public String  iName;           // index file name
	public String  rName;           // reference name, used in ACE
	public boolean read;            // open for read
	public int     dfn;             // data file number
	public int     dfns;            // data file number
	public long    oPointer;        // oligo Pointer in total
	public Vector<Oligo> bunkPet[][]; // [side][rc]
	// public Vector<PET> pets;
	// public int     nopet;
	
	// meta data parameters
	private boolean ow;
	public int    numberOfBACs;
	public String[] BACs;
	public long   minID;
	public long   numberOfOligos;
	public int    numberOfFiles;
	public long   numberOfUQ;
	public long   numberOfUD;
	public double weightOfUM;
	public long   countUM;
	public double weightOfMM;
	public long   countMM;
	
	BufferedWriter iWriter;
	boolean        iOpen;
	String         key;
	
	public int sortCounter;
	int[] l;
	long idx;
	
	Vector<Oligo> bunk;
	PriorityQueue<OligoFileTop> pQueue;
	OligoFile trg;
	
	// FileInputStream fi; 
	// BufferedInputStream bi;
	DataInputStream strIn;
	
	FileOutputStream fw;
	BufferedOutputStream bw;
	DataOutputStream strOut;
	
	FileInputStream[]     sfi; 
	BufferedInputStream[] sbi;
	DataInputStream[]  sstrIn;

	public class OligoFileTop implements Comparable<OligoFileTop>{
		// public boolean ok;
		public Oligo ol;
		public int   id;
		
		public OligoFileTop() throws Exception {
			ol = new Oligo();
			id = -1;
			// ok = true;
		}
		
		public int compareTo(OligoFileTop oft) {
			return ol.compareTo(oft.ol);
		}
		
		public void loadOligo(Oligo o) {
			for (int i = 0; i < Parameters.now; i++) {
				ol.d[i]   = o.d[i];
			}
			// ol.d[1]   = o.d[1];
			ol.ch     = o.ch;
			ol.start  = o.start;
			ol.id1    = o.id1;
			ol.size   = o.size;
			ol.flg    = o.flg;
			ol.weight = o.weight;
	    	ol.u0     = o.u0;	    	
		}
	}

    static void sort_file(String fin, int st) throws Exception { 
    	int oldSt = Oligo.sortIndex;
    	Oligo.sortIndex = st;
    	String fout = fin + ".s";
    	if (Oligo.sortIndex == Oligo.SEQ) {
    		fout = fout + ".sq";
    	} else if (Oligo.sortIndex == Oligo.ID) {
    		fout = fout + ".id";
    	} else if (Oligo.sortIndex == Oligo.WEIGHT) {
    		fout = fout + ".w";
    	} else if (Oligo.sortIndex == Oligo.LOCATION) {
    		fout = fout + ".l";
    	}
		OligoFile ofile = new OligoFile(fin, "", OligoFile.R);
		if (ofile.numberOfOligos > 0) {
			ofile.sort(fout);
		}
		ofile.close();
		ofile = null;
    	Oligo.sortIndex = oldSt;
    }

	public OligoFile() throws Exception {
		// System.err.println("Error: must use a String constructore to create directory\n object not valid");
		vName         = Parameters.pathDB + Parameters.DATABASE;
    	iOpen         = false;
		iName         = vName + "/index.txt";
		rName         = null;
		dfn           = iDfn;
		dfns          = iDfns;
		oPointer      = 0;
		numberOfFiles = 1;
		
		sortCounter   = 0;
		l             = new int[2];
		// nopet         = 0;
    	bunk          = new Vector<Oligo>();
		read          = true;
		idx           = 0;
	}	
	
	public OligoFile(String fName, String refName, int fType) throws Exception {
    	iOpen = false;
		vName         = fName;
		iName         = vName + "/index.txt";
		rName         = refName;
		dfn           = iDfn;
		dfns          = iDfns;
		oPointer      = 0;
		numberOfFiles = 1;
		
		sortCounter   = 0;
		l = new int[2];
		// nopet = 0;
    	bunk = new Vector<Oligo>();

    	if (fType == R) {	
			openRead();         // 151
		} else {
			if(!(new File(vName)).mkdir()) {
				System.err.println("Faile to create the directory " + vName);
			}
			openWrite();
		}
	}

	//*   this constructer is used by JoinOl.java class
	public OligoFile(String fullName) throws Exception {
		int f;
		int k;
		int ID;
		int next;
		boolean EOF;
		Byte B1;
		short l1;
    	iOpen = false;
		String outOl  = Parameters.pathData + "s.ol0/";
		String outPrb = Parameters.pathData + "/s.prb";
    	// System.err.println(outPrb);
		FileOutputStream osPrb    = new FileOutputStream(outPrb);
		DataOutputStream dataout2 = new DataOutputStream(new BufferedOutputStream(osPrb, Parameters.buffSize));
		DataInputStream strInPrb;
		OligoFile[] inOF;
		int numberOfDirs = 0;
		String[] inFiles;
		File fl = new File(Parameters.pathParallel);
		File[] dirs = fl.listFiles();

		for (int dd = 0; dd < dirs.length; dd++) {
			if (dirs[dd].isDirectory()) {
				numberOfDirs++;
			}
		}
		inFiles = new String[numberOfDirs];
		f = 0;
		for (int dd = 0; dd < dirs.length; dd++) {
			if (dirs[dd].isDirectory()) {
				inFiles[f] = new String(dirs[dd]+ "");
				f++;
			}
		}
		inOF  = new OligoFile[numberOfDirs];
		int[][] shift_array = new int[numberOfDirs + 1][3];
		
		for (int i = 0; i < inFiles.length; i++) {
			// System.err.println(i + "\t" + inFiles[i] + "/s.prb");
			strInPrb = new DataInputStream(new BufferedInputStream(new FileInputStream(inFiles[i] + "/s.prb"), Parameters.buffSize)) ;
			k = 0;
			ID = 0;
			EOF = false;
			while (!EOF) {
				try {
					ID = strInPrb.readInt();
					if (k == 0) {
						if (i == 0) {
							shift_array[i][0] = ID;
							shift_array[i][2] = 0;
						} else {
							shift_array[i][0] = shift_array[i - 1][1] - 1;
							shift_array[i][2] = ID - shift_array[i - 1][1] + 1;
						}
						dataout2.writeInt(shift_array[i][0]);
					} else {
						if (i == 0) {
							dataout2.writeInt(ID);
						} else {
							dataout2.writeInt(ID - shift_array[i][2]);
						}
					}

					B1 = strInPrb.readByte();
					l1 = strInPrb.readShort();
					dataout2.writeByte(B1);
					dataout2.writeShort(l1);
					for (int ii = 0; ii < l1; ii++) {
						dataout2.writeInt(strInPrb.readInt());
						/*
						for (int jj = 0; jj < 4; jj++) { // can read and save one int
							Byte B2 = strInPrb.readByte();
							dataout2.writeByte(B2);
						} //*/
					}
					k++;
				} catch (EOFException e) {
					EOF = true;
				}
			}
			if (i == 0) {
				shift_array[i][1] = ID;
			} else {
				shift_array[i][1] = ID - shift_array[i][2];
			}
			// System.err.println(shift_array[i][0] + "\t" + shift_array[i][1] + "\t" + shift_array[i][2]);
			strInPrb.close();
		} // ===============================================
		dataout2.close();

		// [numberOfDirs + 1][3]
		/*
		for (int i = 0; i < numberOfDirs; i++) {
			for (int j = 0; j < 3; j++) {
				System.err.print(shift_array[i][j] + "\t");
			}
			System.err.println();
		}
		//*/
		//*
		OligoFile CompleteOligoFile = new OligoFile(outOl, "", OligoFile.W);

		PriorityQueue<OligoFileTop> pQueue = new PriorityQueue<OligoFileTop>();
		OligoFileTop oft;
		Oligo.sortIndex = Oligo.SEQ;
		for (int i = 0 ; i < inOF.length; i++) {
			inOF[i] = new OligoFile(inFiles[i], "", OligoFile.R);
			oft = new OligoFileTop();
			Oligo ol = new Oligo();
			inOF[i].readOligo(ol);
			oft.ol = ol;
			oft.id = i;
			Oligo.sortIndex = Oligo.SEQ;
			pQueue.add(oft);
		}
		
		while (pQueue.size() > 0) {
			Oligo.sortIndex = Oligo.SEQ;
			oft = pQueue.poll();
			next = oft.id;
			if (oft.id != 0) {
				oft.ol.id1 = oft.ol.id1 - shift_array[oft.id][2];
			}
			oft.ol.start = -1;
			CompleteOligoFile.writeOligo(oft.ol);
			Oligo ol = new Oligo();
			if (inOF[next].readOligo(ol)) {	
				oft.ol = ol;
				oft.id = next;
				Oligo.sortIndex = Oligo.SEQ;
				pQueue.add(oft);
			}
		}

		CompleteOligoFile.close();
		CompleteOligoFile.writeMeta();
		//*
	}		
	//*/
	
    public void finalize() throws Exception {
	}
	public void close() throws Exception {
    	if (read) {
    		if (idx > 0)
      			closeFileRead();
    	} else {
    		// key = "ow";
    		if(ow)
    			closeFileWrite();
    		if (iOpen)
    			iWriter.close();
    		iOpen = false;
    		writeMeta();
    	}
    	bunk.clear();
    	bunk.trimToSize();
    	System.gc();
	}
    
	public void closeP() throws Exception {
		// System.err.println(vName + "\tclossingP");
    	if (read) {
    		if (idx > 0)
      			closeFileRead();
    	} else {
    		// key = "ow";
    		if(ow)
    			closeFileWrite();
    		if (iOpen)
    			iWriter.close();
    		iOpen = false;
    		writeMetaP();
    	}
    	bunk.clear();
    	bunk.trimToSize();
    	System.gc();
	}
    
	public void openRead() throws Exception {
		read = true;
		idx  = 0;
		readMeta(); // 305
	}
	public void openWrite() throws Exception {
		read = false;
		ow   = false;
		clearMeta();
		iWriter = new BufferedWriter(new FileWriter(vName + "/index.txt"));
		key = "Open";
    	iOpen = true;
	}

    public void readMeta() throws Exception {
		StringTokenizer tokens;
		String line;
		BufferedReader reader = new BufferedReader(new FileReader(vName + "/meta.txt")); // 317
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);	
		if (tokens.countTokens() >= 2) {
			tokens.nextToken();
			rName      = tokens.nextToken();
		}
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		if (tokens.countTokens() > 2)
			prbLoaded = true;
		else
			prbLoaded = false;
		tokens.nextToken();
		minID          = new Long(tokens.nextToken());
		// System.err.println("prbLoaded\t" + prbLoaded);
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		numberOfOligos = new Long(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		numberOfUQ     = new Long(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		numberOfUD     = new Long(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		weightOfUM     = new Double(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		weightOfMM     = new Double(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);
		tokens.nextToken();
		numberOfFiles  = new Integer(tokens.nextToken());
		
		line           = reader.readLine();
		tokens         = new StringTokenizer(line);	
		tokens.nextToken();
		numberOfBACs   = new Integer(tokens.nextToken());

		if (numberOfBACs > 0) {
			BACs = new String[numberOfBACs];
			for (int i = 0; i < numberOfBACs; i++) {
				BACs[i] = reader.readLine();
			}
		}
		reader.close();
    }  
    public void writeMeta() throws Exception {
		// System.err.println(vName + "\twriting meta");
    	BufferedWriter writer = new BufferedWriter(new FileWriter(vName + "/meta.txt"));
    	// System.err.println(vName + "/meta.txt,\tminID\t" + (minID-1));
    	writer.write("Ref:\t" + rName); // numberOfIDs
    	writer.write("\nminID:\t"      + minID);
    	writer.write("\nnumberOfOligos:\t" + numberOfOligos);
    	writer.write("\nnumberOfUQ:\t" + numberOfUQ);
    	writer.write("\nnumberOfUD:\t" + numberOfUD);
    	if (countUM > 0)
    		writer.write("\nweightOfUM:\t" + weightOfUM / countUM);
    	else 
        	writer.write("\nweightOfUM:\t" + weightOfUM);
    	if (countMM > 0)
    		writer.write("\nweightOfMM:\t" + weightOfMM / countMM);
    	else
    		writer.write("\nweightOfMM:\t" + weightOfMM);
    	writer.write("\ndfn:\t" + dfn);
    	writer.write("\nnumberOfBACs:\t" + numberOfBACs + "\n");
		for (int i = 0; i < numberOfBACs; i++) {
			writer.write(BACs[i] + "\n");
		}
    	writer.close();
    }

    public void writeMetaP() throws Exception {
		// System.err.println(vName + "\twriting metaP");
    	BufferedWriter writer = new BufferedWriter(new FileWriter(vName + "/meta.txt"));
    	// System.err.println(vName + "/meta.txt,\tminID\t" + (minID-1));
    	writer.write("Ref:\t" + rName); // numberOfIDs
    	writer.write("\nminID:\t"      + minID + "\tPRB");
    	writer.write("\nnumberOfOligos:\t" + numberOfOligos);
    	writer.write("\nnumberOfUQ:\t" + numberOfUQ);
    	writer.write("\nnumberOfUD:\t" + numberOfUD);
    	if (countUM > 0)
    		writer.write("\nweightOfUM:\t" + weightOfUM / countUM);
    	else 
        	writer.write("\nweightOfUM:\t" + weightOfUM);
    	if (countMM > 0)
    		writer.write("\nweightOfMM:\t" + weightOfMM / countMM);
    	else
    		writer.write("\nweightOfMM:\t" + weightOfMM);
    	writer.write("\ndfn:\t" + dfn);
    	writer.write("\nnumberOfBACs:\t" + numberOfBACs + "\n");
		for (int i = 0; i < numberOfBACs; i++) {
			writer.write(BACs[i] + "\n");
		}
    	writer.close();
    }

    public void clearMeta() {
    	minID          = Parameters.maxId;
    	numberOfOligos = 0;
    	weightOfUM     = 0;
    	weightOfMM     = 0;
    	countUM        = 0;
    	countMM        = 0;
    	numberOfUQ     = 0;
    	numberOfUD     = 0;
    	numberOfFiles  = 0;
    }
    public void displayMeta() {
		System.out.println( "\n\n" + 
				"Meta data for\t" + vName + "\n" +
				"Number of files .................:\t" + numberOfFiles + "\n" +
				"Number of Oligos ................:\t" + numberOfOligos + "\n" +
				"Minimum IDs .....................:\t" + minID + "\n" + 
				"Weight of Oligos ................:\t" + (long) (weightOfUM  + weightOfMM) + "\n" + 
				"Average Oligo Weight ............:\t" + Parameters.trim((weightOfUM  + weightOfMM) / (Parameters.maxId - minID)) + "\n" + 
				"Number of UQ ....................:\t" + numberOfUQ + "\n" + 
				"Number of UD ....................:\t" + numberOfUD + "\n" + 
				"Multiplication ratio of UD Oligos:\t" + ((float) numberOfUD / ((float) ((Parameters.maxId - minID) - numberOfUQ))) + "\n" +
				"Weight of UM ....................:\t" + (long) weightOfUM + "\n" + 
				"Weight of MM ....................:\t" + (long) weightOfMM + "\n" +
				"Ratio Weight UM/MM ..............:\t" + Parameters.trim(((float) weightOfUM / (float) weightOfMM)) +
				"\n");
    }
    
	// count[S0/S1/SAll/SAnd/SOr][ALL/Crisp/LQ/UM/UMCrisp/UMLQ]
    public long info(int fStart, long[][] count, long[] countTotal, long[][] countSize, long[] countMM, long[][] weightHist, long[] whCount) throws Exception {
    	int nid = 0;
    	int wt;
    	if (minID != 0)
    		nid = (int) (Parameters.maxId + 1 - minID);
    	long[] histo = new long[Parameters.dbOligoSize + 1];
    	int hi;
    	byte[][] cnt;
    	byte[][] cntLQ;
    	byte[][] cntCrsp;
    	byte[][] cntUM;
    	byte[][] cntUMLQ;
    	byte[][] cntUMCrsp;
    	int cntId;
		cnt       = new byte[2][(int) nid];
		cntLQ     = new byte[2][(int) nid];
		cntCrsp   = new byte[2][(int) nid];
		cntUM     = new byte[2][(int) nid];
		cntUMLQ   = new byte[2][(int) nid];
		cntUMCrsp = new byte[2][(int) nid];
		for(int ii = 0; ii < nid; ii++) {
			cnt[0][ii]       = cnt[1][ii]       = 0;
			cntLQ[0][ii]     = cntLQ[1][ii]     = 0;
			cntCrsp[0][ii]   = cntCrsp[1][ii]   = 0;
			cntUM[0][ii]     = cntUM[1][ii]     = 0;
			cntUMLQ[0][ii]   = cntUMLQ[1][ii]   = 0;
			cntUMCrsp[0][ii] = cntUMCrsp[1][ii] = 0;
		}
		countTotal[0] = countTotal[1] = idx = 0;
		for (int i = 0; i < 5; i++) {
			for (int j = 0; j < 6; j++) {
				count[i][j] = 0;
			}
		}
		for (int i = 0; i < countSize.length; i++) {
			for(int j = 0; j < countSize[i].length; j++) {
				countSize[i][j] = 0;
			}
		}
    	if (read) {
   			openFileRead(vName + "/file" + fStart);
    		dfn = fStart;
        	Oligo ol = new Oligo();
			idx = 0;
			while(idx < numberOfOligos) { 
				readOligo(ol, "/file", Parameters.fSize);
				wt = (int) (ol.weight * 100);
				if (wt <= 100 && wt >= 0) {
					if(ol.isUM()) {
						weightHist[0][wt]++;
						whCount[0]++;
					} else {
						weightHist[1][wt]++;
						whCount[1]++;
					}
				}
				if(ol.isPET())
					countTotal[1]++;
				countSize[ol.getSide()][ol.size]++;
				countMM[ol.getSide()] += ol.u0;
	   			countTotal[0]++;
				cntId = Parameters.maxId - ol.id1;
				if (ol.u0 >= 0) {
					if (ol.u0 < Parameters.dbOligoSize)
						hi = ol.u0;
					else 
						hi = Parameters.dbOligoSize;
					histo[hi]++;
				} 
				if (cntId < nid) {
					cnt[ol.getSide()][cntId] = 1;
					if (ol.isLQ()) {
						cntLQ[ol.getSide()][cntId] = 1;
					}
					if (ol.isUQ()) {
						cntCrsp[ol.getSide()][cntId] = 1;
					}
					if (ol.isUM()) {
						cntUM[ol.getSide()][cntId] = 1;
						if (ol.isLQ()) {
							cntUMLQ[ol.getSide()][cntId] = 1;
						}
						if (ol.isUQ()) {
							cntUMCrsp[ol.getSide()][cntId] = 1;
						}
					}
				} 

				idx++;
			}		
    	} else {
			System.err.println(" File open for write");
		}
     	cntId = 0;
    	for(int ii = 0; ii < nid; ii++) {
    		if(cnt[0][ii] > 0 || cnt[1][ii] > 0) {
    			count[SOr][ALL]++;
    		}
    		if(cnt[0][ii] > 0) {
    			count[S0][ALL]++;
    			count[SAll][ALL]++;
        		if(cnt[1][ii] > 0) {
        			count[SAnd][ALL]++;
        		}
    		}
    		if(cnt[1][ii] > 0) {
    			count[S1][ALL]++;
    			count[SAll][ALL]++;
    		}
    		
    		if(cntCrsp[0][ii] > 0 || cntCrsp[1][ii] > 0) {
    			count[SOr][Crisp]++;
    		}
    		if(cntCrsp[0][ii] > 0) {
    			count[S0][Crisp]++;
    			count[SAll][Crisp]++;
        		if(cntCrsp[1][ii] > 0) {
        			count[SAnd][Crisp]++;
        		}
    		}
    		if(cntCrsp[1][ii] > 0) {
    			count[S1][Crisp]++;
    			count[SAll][Crisp]++;
    		}

    		if(cntLQ[0][ii] > 0 || cntLQ[1][ii] > 0) {
    			count[SOr][LQ]++;
    		}
    		if(cntLQ[0][ii] > 0) {
    			count[S0][LQ]++;
    			count[SAll][LQ]++;
        		if(cntLQ[1][ii] > 0) {
        			count[SAnd][LQ]++;
        		}
    		}
    		if(cntLQ[1][ii] > 0) {
    			count[S1][LQ]++;
    			count[SAll][LQ]++;
    		}

    		if(cntUM[0][ii] > 0 || cntUM[1][ii] > 0) {
    			count[SOr][UM]++;
    		}
    		if(cntUM[0][ii] > 0) {
    			count[S0][UM]++;
    			count[SAll][UM]++;
        		if(cntUM[1][ii] > 0) {
        			count[SAnd][UM]++;
        		}
    		}
    		if(cntUM[1][ii] > 0) {
    			count[S1][UM]++;
    			count[SAll][UM]++;
    		}

    		if(cntUMCrsp[0][ii] > 0 || cntUMCrsp[1][ii] > 0) {
    			count[SOr][UMCrisp]++;
    		}
    		if(cntUMCrsp[0][ii] > 0) {
    			count[S0][UMCrisp]++;
    			count[SAll][UMCrisp]++;
        		if(cntUMCrsp[1][ii] > 0) {
        			count[SAnd][UMCrisp]++;
        		}
    		}
    		if(cntUMCrsp[1][ii] > 0) {
    			count[S1][UMCrisp]++;
    			count[SAll][UMCrisp]++;
    		}

    		if(cntUMLQ[0][ii] > 0 || cntUMLQ[1][ii] > 0) {
    			count[SOr][UMLQ]++;
    		}
    		if(cntUMLQ[0][ii] > 0) {
    			count[S0][UMLQ]++;
    			count[SAll][UMLQ]++;
        		if(cntUMLQ[1][ii] > 0) {
        			count[SAnd][UMLQ]++;
        		}
    		}
    		if(cntUMLQ[1][ii] > 0) {
    			count[S1][UMLQ]++;
    			count[SAll][UMLQ]++;
    		}
    	}


    	BufferedWriter writer = new BufferedWriter(new FileWriter(vName + "/histogramU0.txt"));
    	for (int i = 0; i < histo.length; i++) {
    		writer.write(i + "\t" + histo[i] + "\n");
    	}
    	writer.close();
    	return idx;
    }
 
    public void readHeadId(int t, int fStart, int id) throws Exception {
    	int oCount = 0;
		// System.err.println("readHeadId\t" + vName + "\t" + id);
    	if (read) {
    		if (t == OligoFile.F) 
    			openFileRead(vName + "/file" + fStart);
    		else 
    			openFileRead(vName + "/sort" + fStart);
    		dfn = fStart;
        	Oligo ol = new Oligo();
			idx = 0;
			while(idx < numberOfOligos) { //strIn[DATA].available() > 0) {
				if (t == OligoFile.F) 
					readOligo(ol, "/file", Parameters.fSize);
				else
					readOligo(ol, "/sort", Parameters.fSortSize);
				if (ol.id1 == id)
					System.err.println(idx + "\t" + oCount++ + ":\t" + ol);
				idx++;
			}		
    	} else {
			System.err.println(" File open for write");
		}
		displayMeta();
   }
    public void ace() throws Exception {
		int fSize = 0;
		String line = "";
		System.err.println("vName:\t" + vName + "\trName:\t" + rName);
    	if (read) {
   			openFileRead(vName + "/file" + 1);
   	     	BufferedWriter writer = new BufferedWriter(new FileWriter(vName + "/ace"));
			BufferedReader readerRef = new BufferedReader(new FileReader(rName));
			line = readerRef.readLine();
			StringTokenizer tokens = new StringTokenizer(line, ":");	
			if (tokens.countTokens() >= 6) {
				// tokens.nextToken();
				for(int i = 0; i < 5; i++)
					tokens.nextToken();
				fSize = new Integer(tokens.nextToken());
			}
			// System.err.println(line + "\t" + fSize);

   	     	writer.write("AS 1 " + numberOfOligos + "\n");
   	     	writer.write("CO " + rName + " " + fSize + " " + numberOfOligos + " U\n");
    		dfn = 1;
        	Oligo ol = new Oligo();
			idx = 0;
			while(idx < numberOfOligos) { //strIn[DATA].available() > 0) {
				readOligo(ol, "/file", Parameters.fSize);
				writer.write(ol.start + "\t" + ol.toSeq() + "\n");
				idx++;
			}		
			readerRef.close();
			writer.close();
    	} else {
			System.err.println(vName + " File open for write");
		}
		// displayMeta();
   }

    public void readHead(int last, int first) throws Exception {
    	if (read) {
   			openFileRead(vName + "/file" + 1);
    		dfn = 1;
        	Oligo ol = new Oligo();
			idx = 0;
			// System.out.println("readHead:\t");
			while(idx < numberOfOligos) { //strIn[DATA].available() > 0) {
				readOligo(ol, "/file", Parameters.fSize);
				if ((idx >= first && idx < last) || last < 0)
					System.out.println(idx + ":\t" + ol);
				if (idx > last && last > 0) 
					break;
				idx++;
			}		
    	} else {
			System.err.println(" File open for write");
		}
		// displayMeta();
   }
    
    public long available() {
    	return numberOfOligos - idx;
    }
    
    public boolean readOligo(Oligo ol) throws Exception {
    	boolean ret = true;
    	if (idx == 0) {
    		dfn = 1;
    		// System.err.println("Opening file:\t" + dfn);
			openFileRead(vName + "/file" + dfn);
    	}
    	if (idx < numberOfOligos) {
    		ret = readOligo(ol, "/file", Parameters.fSize);
    	} else {
    		ret = false;
    	}
    	idx++;
    	return ret;
    }
    public boolean readOligo(Oligo ol, String fil, int fsz) throws Exception {
    	boolean ret = true;
    	int fl = (int) (1 + oPointer / (long) fsz + (long) iDfn);
    	if (fl > dfn) {
  			closeFileRead();
			dfn++;
			try {
				// System.err.println("Opening file:\t" + dfn);
				openFileRead(vName + fil + dfn); // ----------------------------------------------------- "/file"
			} catch (Exception e) {
				System.err.println("Error opening file:\t" + dfn + "\t" + e);
			}
    	}
    	try {
    		for (int i = 0; i < Parameters.now; i++) {
    			ol.d[i]  = strIn.readLong();
    		}
        	ol.ch    = strIn.readByte();
        	ol.start = strIn.readInt();
        	ol.id1   = strIn.readInt();
        	ol.size  = strIn.readByte();
        	ol.flg   = strIn.readByte();
        	ol.u0    = strIn.readByte();
        	ol.weight = strIn.readByte();
        	if (prbLoaded) {
        		ol.prbSize = strIn.readShort();
        		ol.petSize = strIn.readShort();
        		// System.err.println("R <" + ol.petSize + "\t" + ol.prbSize + ">");
        		ol.prb = new byte[4][ol.prbSize];
            	for (int ii = 0; ii < ol.prbSize; ii++) {
            		for (int b = 0; b < 4; b++) {
            			ol.prb[b][ii] = strIn.readByte();
            		}
            	}
        	}
    	} catch (Exception e) {
    		ret = false;
    		System.err.println("oPointer:\t" + oPointer + "\t" + e);
    	}
    	ol.weight = ol.weight / 100;
    	oPointer++;
    	return ret;
    }
    
    public void writeOligo(Oligo ol) throws Exception {
    	if (ol.id1 < minID)
			minID = ol.id1;
    	if (ol.isUM()) {
    		weightOfUM += ol.weight;
    		countUM++;
    	} else {
    		weightOfMM += ol.weight;
    		countMM++;
    	}
		if (ol.id1 < minID)
			minID = ol.id1;
		if (ol.isUQ())
			numberOfUQ += 1;
		else 
			numberOfUD += 1;
    	if (numberOfOligos % Parameters.fSize == 0) {
    		// if (numberOfOligos > 0)
    		
    		if (dfn > iDfn) {
    			saveKey(ol);
    			closeFileWrite();
    		}
			dfn++;
			openFileWrite(vName + "/file" + dfn);
    	}
    	for (int i = 0; i < Parameters.now; i++) {
    		strOut.writeLong(ol.d[i]);
    	}
    	strOut.writeByte(ol.ch); // int
    	strOut.writeInt(ol.start);
    	strOut.writeInt(ol.id1);
    	strOut.writeByte(ol.size); // int
    	strOut.writeByte(ol.flg); // int
   		strOut.writeByte(ol.u0); // int
    	strOut.writeByte((byte) Math.round(100 * ol.weight)); // float
    	numberOfOligos++;
    }

    public void writeOligoP(Oligo ol) throws Exception {
    	if (ol.id1 < minID)
			minID = ol.id1;
    	if (ol.isUM()) {
    		weightOfUM += ol.weight;
    		countUM++;
    	} else {
    		weightOfMM += ol.weight;
    		countMM++;
    	}
		if (ol.id1 < minID)
			minID = ol.id1;
		if (ol.isUQ())
			numberOfUQ += 1;
		else 
			numberOfUD += 1;
    	if (numberOfOligos % Parameters.fSize == 0) {
    		// if (numberOfOligos > 0)
    		
    		if (dfn > iDfn) {
    			saveKey(ol);
    			closeFileWrite();
    		}
			dfn++;
			openFileWrite(vName + "/file" + dfn);
    	}
    	for (int i = 0; i < Parameters.now; i++) {
    		strOut.writeLong(ol.d[i]);
    	}
    	strOut.writeByte(ol.ch); // int
    	strOut.writeInt(ol.start);
    	strOut.writeInt(ol.id1);
    	strOut.writeByte(ol.size); // int
    	strOut.writeByte(ol.flg); // int
   		strOut.writeByte(ol.u0); // int
    	strOut.writeByte((byte) Math.round(100 * ol.weight)); // float
    	strOut.writeShort(ol.prbSize);
    	strOut.writeShort(ol.petSize);
    	for (int ii = 0; ii < ol.prbSize; ii++) {
    		for (int b = 0; b < 4; b++) {
    			strOut.writeByte(ol.prb[b][ii]);
    		}
    	}
    	numberOfOligos++;
    }

    public void writeSOligo(Oligo ol) throws Exception {
    	if (numberOfOligos % Parameters.fSortSize == 0) {
    		if (dfns > iDfns) {
    			saveKey(ol);
    			closeFileWrite();
    		}
			dfns++;
			openFileWrite(vName + "/sort" + dfns);
    	}
		if (ol.id1 < minID)
			minID = ol.id1;
		for (int i = 0; i < Parameters.now; i++) {
			strOut.writeLong(ol.d[i]);
		}
    	strOut.writeByte(ol.ch); // int
    	strOut.writeInt(ol.start);
    	strOut.writeInt(ol.id1);
    	strOut.writeByte(ol.size); // int
    	strOut.writeByte(ol.flg); // int
   		strOut.writeByte(ol.u0); // int
    	strOut.writeByte((byte) Math.round(100 * ol.weight)); // float
    	numberOfOligos++;
    }

    public void writeSOligoP(Oligo ol) throws Exception {
    	if (numberOfOligos % Parameters.fSortSize == 0) {
    		if (dfns > iDfns) {
    			saveKey(ol);
    			closeFileWrite();
    		}
			dfns++;
			openFileWrite(vName + "/sort" + dfns);
    	}
		if (ol.id1 < minID)
			minID = ol.id1;
		for (int i = 0; i < Parameters.now; i++) {
			strOut.writeLong(ol.d[i]);
		}
    	strOut.writeByte(ol.ch); 
    	strOut.writeInt(ol.start);
    	strOut.writeInt(ol.id1);
    	strOut.writeByte(ol.size); 
    	strOut.writeByte(ol.flg); 
   		strOut.writeByte(ol.u0); 
    	strOut.writeByte((byte) Math.round(100 * ol.weight)); 
    	strOut.writeShort(ol.prbSize);
    	strOut.writeShort(ol.petSize);
    	for (int ii = 0; ii < ol.prbSize; ii++) {
    		for (int b = 0; b < 4; b++) {
    			strOut.writeByte(ol.prb[b][ii]);
    		}
    	}
    	numberOfOligos++;
    }

    public void sort_WriteOligo(Oligo ol, int sortOn) throws Exception {
    	if (ol.id1 < minID)
			minID = ol.id1;
    	if (ol.isUM()) {
    		weightOfUM += ol.weight;
    		countUM++;
    	} else {
    		weightOfMM += ol.weight;
    		countMM++;
    	}
		if (ol.isUQ())
			numberOfUQ += 1;
		else 
			numberOfUD += 1;
		bunk.add(ol);//(new Oligo(ol));
    	if (bunk.size() >= Parameters.fSortSize) {
    		saveSBunk(sortOn);
    	}
    }
    
    public void sort_WriteOligoP(Oligo ol, int sortOn) throws Exception {
    	if (ol.id1 < minID)
			minID = ol.id1;
    	if (ol.isUM()) {
    		weightOfUM += ol.weight;
    		countUM++;
    	} else {
    		weightOfMM += ol.weight;
    		countMM++;
    	}
		if (ol.isUQ())
			numberOfUQ += 1;
		else 
			numberOfUD += 1;
		bunk.add(ol);//(new Oligo(ol));
    	if (bunk.size() >= Parameters.fSortSize) {
    		saveSBunkP(sortOn);
    	}
    }

    public void saveSBunk(int sIndex) throws Exception {
    	int bk = Oligo.sortIndex;
    	Oligo ol;
    	Oligo.sortIndex = sIndex;
		Collections.sort(bunk);
	    Iterator<Oligo> it = bunk.iterator();
        for (; it.hasNext(); ) {
        	ol = (Oligo) it.next ();
        	writeSOligo(ol);
        }
		bunk.clear();
		Oligo.sortIndex = bk;
    }

    public void saveSBunkP(int sIndex) throws Exception {
    	int bk = Oligo.sortIndex;
    	Oligo ol;
    	Oligo.sortIndex = sIndex;
		Collections.sort(bunk);
	    Iterator<Oligo> it = bunk.iterator();
        for (; it.hasNext(); ) {
        	ol = (Oligo) it.next ();
        	// System.err.println(ol + "\t" + ol.prbSize + "\t" + ol.prb);
        	writeSOligoP(ol);
        }
		bunk.clear();
		System.gc();
		Oligo.sortIndex = bk;
    }
    //*/
    public void openFileWrite(String fl) throws Exception {
    	// System.err.println("openFileWrite" + fl);
    	ow = true;
		fw     = new FileOutputStream(fl);
	    bw     = new BufferedOutputStream(fw, Parameters.buffSize);
	    strOut = new DataOutputStream(bw);	    		
    }
    private void closeFileWrite() throws Exception {
    	strOut.flush();
    	bw.flush();
    	fw.flush();
    	bw.flush();  	
    	strOut.flush();
    	
		strOut.close();    	
		bw.close();
		fw.close();
    	System.err.print(" *");
    	if (iOpen) {
    		iWriter.write(key + "\n");
    		iWriter.flush();
    	}
    }
    private void openFileRead(String fl) throws Exception {
		// System.err.println("\nSorting file " + dfn + "\nTiming:\tRead\tSort\tWrite\tFile");
    	if(numberOfOligos > 0) {
    		strIn = new DataInputStream(new BufferedInputStream(new FileInputStream(fl), Parameters.buffSize));
    	}
    }
    private void closeFileRead() throws Exception {
		strIn.close();    	
    }

    public  void join(String[] lst, int sIndex) throws Exception {
    	boolean u0 = true;
    	int count = 0;
    	Oligo ol = new Oligo();
		bunk.clear();
    	for (int i = 0; i < lst.length; i++) {
    		OligoFile fl = new OligoFile(lst[i], "", OligoFile.R);
    		fl.dfn = iDfn + 1;
    		fl.openFileRead(fl.vName + "/file" + fl.dfn);
    		while(fl.strIn.available() > 0 || fl.dfn < numberOfFiles) {
        		fl.readOligo(ol, "/file", Parameters.fSize); // 332
        		if (i == 0)
        			u0 = true;
        		else
        			u0 = false;
        		ol.setU0(u0);
        		bunk.add(new Oligo(ol));
        		count++;
    	    	if (count >= Parameters.fSortSize) {
    	    		saveSBunk(sIndex);
    	    		count = 0;
    	    	}
    		}
    		fl.close();
    	}
		// Collections.sort(trg.bunk);
		saveSBunk(sIndex);
		numberOfFiles = dfn        = dfns;
    }
    
    public  void sortDB() throws Exception {
    	Oligo.sortIndex = Oligo.SEQ;
    	sort1_DB(vName);
    	trg  = new OligoFile(vName, "", OligoFile.R);
    	trg.joinSort();
    	trg.close();
    }
    public  void sort(String targetFile) throws Exception {
    	sort1(targetFile);
    	trg  = new OligoFile(targetFile, rName, OligoFile.R);
    	trg.joinSort();
    	trg.close();
    }
    private void saveBunk() throws Exception { // used by sort1
    	Oligo ol;
	    Iterator<Oligo> it = bunk.iterator();
        while (it.hasNext()) {
        	ol = (Oligo) it.next ();
        	trg.writeSOligo(ol);
        }
		bunk.clear();
		System.gc();
    }
//*
    private void sort1(String targetFile) throws Exception {
    	int count = 0;
    	Oligo ol = new Oligo();

    	trg = new OligoFile(targetFile, rName, OligoFile.W);
    	bunk.clear();
		dfn = iDfn + 1;
		openFileRead(vName + "/file" + dfn);
    	
		while(strIn.available() > 0 || dfn < numberOfFiles) {
    		readOligo(ol, "/file", Parameters.fSize); // 332
	    	bunk.add(new Oligo(ol));
    		count++;
	    	if (count >= Parameters.fSortSize) {
	    		Collections.sort(bunk);
		    	saveBunk();
	    		count = 0;
	    	}
		}
		Collections.sort(bunk);
    	saveBunk();
    	trg.dfn        = trg.dfns;
    	trg.numberOfUD = numberOfUD;
    	trg.numberOfUQ = numberOfUQ;
    	trg.weightOfUM = weightOfUM;
    	trg.weightOfMM = weightOfMM;
    	trg.countUM    = countUM;
    	trg.countMM    = countMM;
    	trg.close();
    } //*/

    private void sort1_DB(String targetFile) throws Exception {
    	int count = 0;
    	Oligo ol = new Oligo();

    	trg = new OligoFile(targetFile, rName, OligoFile.W);
    	bunk.clear();
		dfn = iDfn + 1;
		openFileRead(vName + "/file" + dfn);
    	
		while(strIn.available() > 0 || dfn < numberOfFiles) {
    		readOligo(ol, "/file", Parameters.fSize); // 332
    		if(ol.isRC())
    			ol.revers();
	    	bunk.add(new Oligo(ol));
    		count++;
	    	if (count >= Parameters.fSortSize) {
	    		Collections.sort(bunk);
		    	saveBunk();
	    		count = 0;
	    	}
		}
		if(count > 0) {
			Collections.sort(bunk);
			saveBunk();
		}
    	trg.dfn        = trg.dfns;
    	trg.numberOfUD = numberOfUD;
    	trg.numberOfUQ = numberOfUQ;
    	trg.weightOfUM = weightOfUM;
    	trg.weightOfMM = weightOfMM;
    	trg.countUM    = countUM;
    	trg.countMM    = countMM;
    	trg.close();
    } //*/
    private boolean getOligo(Oligo ol, int frm) throws Exception { // used by joinSort
    	boolean ret = true;
    	if (sstrIn[frm].available() > 0) {
    		for (int i = 0; i < Parameters.now; i++) {
    			ol.d[i]  = sstrIn[frm].readLong();
    		}
        	ol.ch    = sstrIn[frm].readByte();
        	ol.start = sstrIn[frm].readInt();
        	ol.id1   = sstrIn[frm].readInt();
        	ol.size  = sstrIn[frm].readByte();
        	ol.flg   = sstrIn[frm].readByte();
       		ol.u0    = sstrIn[frm].readByte();
        	ol.weight = sstrIn[frm].readByte();
        	ol.weight = ol.weight / 100;
    	} else {
    		ret = false;
    	}
    	return ret;
    }
    
    private boolean getOligoP(Oligo ol, int frm) throws Exception { // used by joinSort
    	boolean ret = true;
    	if (sstrIn[frm].available() > 0) {
    		for (int i = 0; i < Parameters.now; i++) {
    			ol.d[i]  = sstrIn[frm].readLong();
    		}
        	ol.ch    = sstrIn[frm].readByte();
        	ol.start = sstrIn[frm].readInt();
        	ol.id1   = sstrIn[frm].readInt();
        	ol.size  = sstrIn[frm].readByte();
        	ol.flg   = sstrIn[frm].readByte();
       		ol.u0    = sstrIn[frm].readByte();
        	ol.weight = sstrIn[frm].readByte();
        	ol.weight = ol.weight / 100;
        	ol.prbSize = sstrIn[frm].readShort();
        	ol.petSize = sstrIn[frm].readShort();
    		// System.err.println("S <" + ol.petSize + "\t" + ol.prbSize + ">");
        	ol.prb = new byte[4][ol.prbSize];
        	for (int ii = 0; ii < ol.prbSize; ii++) {
        		for (int b = 0; b < 4; b++) {
        			ol.prb[b][ii] = sstrIn[frm].readByte();
        		}
        	}

    	} else {
    		ret = false;
    	}
    	return ret;
    }

    public void joinFilterContaminationSort() throws Exception {
    	int noFiles;
    	int countNoise = 0;
    	boolean ntrue = true;
    	int cmp = 0;
    	int next = 0;
    	Oligo noise = new Oligo();
		String noiseDB = Parameters.pathData + "/contaminationDB";
		OligoFile ofNoise = new OligoFile(noiseDB, "", OligoFile.R);                        // 1163
		System.err.println("Contamination database size:\t" + ofNoise.numberOfOligos);
    	noFiles = numberOfFiles;
    	clearMeta();
    	System.err.println("noFiles:\t" + noFiles);
    	pQueue = new PriorityQueue<OligoFileTop>();

    	OligoFileTop oft;
    	
    	sfi = new FileInputStream[noFiles]; 
    	sbi = new BufferedInputStream[noFiles];
    	sstrIn = new DataInputStream[noFiles];
    	
    	dfn = iDfn;
		for (int i = 0; i < noFiles; i++) {
    		oft       = new OligoFileTop();
    		sfi[i]    = new FileInputStream(vName + "/sort"+ (i+1));
    		sbi[i]    = new BufferedInputStream(sfi[i], 5000);
    		sstrIn[i] = new DataInputStream(sbi[i]);
    		getOligo(oft.ol, i); 
    		oft.id = i;
    		pQueue.add(oft);
    	}
		ntrue = ofNoise.readOligo(noise); // =================================================================
		// System.err.println(" ------------------- NOISE:\t" + noise);
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			if(ntrue) {
				cmp = noise.equal(oft.ol, oft.ol.size);
				if(cmp == 0) {
					oft.ol.setLQ(true);
					countNoise++;
					// System.err.println(" ------------------- NOISE:\t" + oft.ol);
				} else if(cmp > 0) {
					while (ntrue && cmp > 0) {
						ntrue = ofNoise.readOligo(noise);
						if (ntrue)
							cmp = noise.equal(oft.ol, oft.ol.size);
					}
				}
			}
			writeOligo(oft.ol);
			if(getOligo(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		closeFileWrite();
    	for (int i = 0; i < noFiles; i++) {
    		sstrIn[i].close();
    		sbi[i].close();
    		sfi[i].close();
    	}
    	writeMeta();
    	ofNoise.close();
    	for (int i = 0; i < noFiles; i++) {
    		(new File(vName + "/sort"+ (i+1))).delete();
    	}
    	System.err.println("countNoise:\t" + countNoise);
    }
    public void joinSortP() throws Exception {
    	int noFiles;
    	int next = 0;
    	noFiles = numberOfFiles;
    	clearMeta();
    	System.err.println(vName + "\njoin noFiles:\t" + noFiles);
    	pQueue = new PriorityQueue<OligoFileTop>();

    	OligoFileTop oft = null;
		iWriter = new BufferedWriter(new FileWriter(vName + "/index.txt"));
		key = "Open";
    	iOpen = true;
    	
    	sfi = new FileInputStream[noFiles]; 
    	sbi = new BufferedInputStream[noFiles];
    	sstrIn = new DataInputStream[noFiles];
    	
    	dfn = iDfn;
		for (int i = 0; i < noFiles; i++) {
    		oft       = new OligoFileTop();
    		sfi[i]    = new FileInputStream(vName + "/sort"+ (i+1));
    		sbi[i]    = new BufferedInputStream(sfi[i], 51200);
    		sstrIn[i] = new DataInputStream(sbi[i]);
    		getOligoP(oft.ol, i); 
    		oft.id = i;
    		pQueue.add(oft);
    	}
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			writeOligoP(oft.ol);
			if(getOligoP(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		// System.err.println("toSave Key");
		saveKey(oft.ol);
		closeFileWrite();
    	for (int i = 0; i < noFiles; i++) {
    		sstrIn[i].close();
    		sbi[i].close();
    		sfi[i].close();
    	}
    	writeMetaP();
    	for (int i = 0; i < noFiles; i++) {
    		(new File(vName + "/sort"+ (i+1))).delete();
    	}
		iWriter.close();
    	iOpen = false;
    }

    public void joinSort() throws Exception {
    	int noFiles;
    	int next = 0;
    	noFiles = numberOfFiles;
    	clearMeta();
    	System.err.println("\nnoFiles:\t" + noFiles);
    	pQueue = new PriorityQueue<OligoFileTop>();

    	OligoFileTop oft = null;
		iWriter = new BufferedWriter(new FileWriter(vName + "/index.txt"));
		key = "Open";
    	iOpen = true;
    	
    	sfi = new FileInputStream[noFiles]; 
    	sbi = new BufferedInputStream[noFiles];
    	sstrIn = new DataInputStream[noFiles];
    	
    	dfn = iDfn;
		for (int i = 0; i < noFiles; i++) {
    		oft       = new OligoFileTop();
    		sfi[i]    = new FileInputStream(vName + "/sort"+ (i+1));
    		sbi[i]    = new BufferedInputStream(sfi[i], 5000);
    		sstrIn[i] = new DataInputStream(sbi[i]);
    		getOligo(oft.ol, i); 
    		oft.id = i;
    		pQueue.add(oft);
    	}
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			writeOligo(oft.ol);
			if(getOligo(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		// System.err.println("toSave Key");
		saveKey(oft.ol);
		closeFileWrite();
    	for (int i = 0; i < noFiles; i++) {
    		sstrIn[i].close();
    		sbi[i].close();
    		sfi[i].close();
    	}
    	writeMeta();
    	for (int i = 0; i < noFiles; i++) {
    		(new File(vName + "/sort"+ (i+1))).delete();
    	}
		iWriter.close();
    	iOpen = false;
    }

    private void saveKey(Oligo o) {
    	if (Oligo.sortIndex == Oligo.ID) {
    		key = "id:\t" + o.id1;
    	} else if (Oligo.sortIndex == Oligo.SEQ) {
    		key = o.toSeq();
    	}  else {
    		key = "x";
    	}
    	// System.err.println("SaveKey:\t" + key);
    }
    
    public void joinSortU2() throws Exception {
    	// int outCount = 0;
    	// Oligo last = new Oligo();
    	int u0;
    	Oligo ol1 = new Oligo(); 
    	l[1] = 0;
    	int noFiles;
    	int next = 0;
    	noFiles = numberOfFiles;

    	clearMeta();
    	// OligoFile out  = new OligoFile(vName + "/out", OligoFile.W);
    	pQueue = new PriorityQueue<OligoFileTop>();

    	OligoFileTop oft = null;
		iWriter = new BufferedWriter(new FileWriter(vName + "/index.txt"));
		key = "Open";
    	iOpen = true;
    	
    	sfi = new FileInputStream[noFiles]; 
    	sbi = new BufferedInputStream[noFiles];
    	sstrIn = new DataInputStream[noFiles];
    	
    	dfn = iDfn;
		for (int i = 0; i < noFiles; i++) {
			System.err.println("sFile\t" + i);
    		oft       = new OligoFileTop();
    		sfi[i]    = new FileInputStream(vName + "/sort"+ (i+1));
    		sbi[i]    = new BufferedInputStream(sfi[i]);
    		sstrIn[i] = new DataInputStream(sbi[i]);
    		getOligo(oft.ol, i); 
    		oft.id = i;
    		pQueue.add(oft);
    	}
		if (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			ol1 = new Oligo(oft.ol);
			if(getOligo(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			l[0] = l[1];
			l[1] = oft.ol.matchR(ol1);
	        if (l[0] > l[1]) {
	        	u0 = l[0];
	        } else {
	        	u0 = l[1];
	        }
    		ol1.u0 = (byte) u0;
    		
			writeOligo(ol1); 
			
	   		ol1 = new Oligo(oft.ol);
    		if(getOligo(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		saveKey(oft.ol);
		closeFileWrite();
    	for (int i = 0; i < noFiles; i++) {
    		sstrIn[i].close();
    		sbi[i].close();
    		sfi[i].close();
    	}
    	writeMeta();
    	for (int i = 0; i < noFiles; i++) {
    		(new File(vName + "/sort"+ (i+1))).delete();
    	}
		iWriter.close();
    	iOpen = false;
    }

    public void joinSortU2DB() throws Exception {
    	int noFiles = 0;
    	int f = 0;
		File fl = new File(vName);
		File[] dirs = fl.listFiles();
		for (int dd = 0; dd < dirs.length; dd++) {
			if(dirs[dd].isDirectory() && (dirs[dd].toString()).indexOf(".db") >= 0) {
		        File[] files = dirs[dd].listFiles();
		        for(int i=0; i<files.length; i++) {
		            if(!files[i].isDirectory() && (files[i].toString()).indexOf("sort") >= 0) {
		            	noFiles++;
		            }
		        }
			}
		}
		// System.err.println("joinSortU2DB\t" + noFiles);
		String[] sortFiles = new String[noFiles];
		for (int dd = 0; dd < dirs.length; dd++) {
			if(dirs[dd].isDirectory()) {
		        File[] files = dirs[dd].listFiles();
		        for(int i=0; i<files.length; i++) {
		            if(!files[i].isDirectory() && (files[i].toString()).indexOf("sort") >= 0) {
		            	sortFiles[f] = new String(files[i] + "");
		            	f++;
		            }
		        }
			}
		}
		/*
		for (int i = 0; i < noFiles; i++) {
			System.err.println(sortFiles[i]);
		}
    	
    	//*/
    	int u0;
    	Oligo ol1 = new Oligo(); 
    	l[1] = 0;
    	int next = 0;

    	clearMeta();
    	pQueue = new PriorityQueue<OligoFileTop>();

    	OligoFileTop oft = null;
		iWriter = new BufferedWriter(new FileWriter(vName + "/index.txt"));
		key = "Open";
    	iOpen = true;
    	
    	sstrIn = new DataInputStream[noFiles];
    	
    	dfn = iDfn;
		for (int i = 0; i < noFiles; i++) {
			// System.err.println("sFile\t" + i);
    		oft       = new OligoFileTop();
    		sstrIn[i] = new DataInputStream(new BufferedInputStream(new FileInputStream(sortFiles[i])));
    		getOligo(oft.ol, i); 
    		oft.id = i;
    		pQueue.add(oft);
    	}
		if (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			ol1 = new Oligo(oft.ol);
			if(getOligo(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		while (pQueue.size() > 0) {
			oft = pQueue.poll();
			next = oft.id;
			l[0] = l[1];
			l[1] = oft.ol.matchR(ol1);
	        if (l[0] > l[1]) {
	        	u0 = l[0];
	        } else {
	        	u0 = l[1];
	        }
    		ol1.u0 = (byte) u0;
    		
			writeOligo(ol1); 
			
	   		ol1 = new Oligo(oft.ol);
    		if(getOligo(oft.ol, next)) {
				oft.id = next;
				pQueue.add(oft);
			}
		}
		saveKey(oft.ol);
		closeFileWrite();
    	for (int i = 0; i < noFiles; i++) {
    		sstrIn[i].close();
    	}
    	writeMeta();
    	for (int i = 0; i < noFiles; i++) {
    		(new File(sortFiles[i])).delete();
    	}
		iWriter.close();
    	iOpen = false;
    	//*/
    }

    public  void fillU1(String targetFile) throws Exception {

     	Oligo ol0 = new Oligo();
    	Oligo ol1 = new Oligo();
    	trg = new OligoFile(targetFile, rName, OligoFile.W);
		dfn = iDfn + 1;
		// System.err.println("Opening file:\t" + dfn);
		openFileRead(vName + "/file" + dfn);
		readOligo(ol0); 
		ol0.u0 = 0;
		trg.writeOligo(ol0);
		while(strIn.available() > 0 || dfn < numberOfFiles) {
    		readOligo(ol1); 
    		ol0.u0 = (byte) ol1.match(ol0);
    		trg.writeOligo(ol1);
    		ol0 = new Oligo(ol1);
		}
		trg.close();
    }

    public  void fillU2(String targetFile) throws Exception {

    	int u0 = 0;
    	Oligo ol0 = new Oligo();
    	Oligo ol1 = new Oligo();
    	l[1] = 0;
    	trg = new OligoFile(targetFile, rName, OligoFile.W);
		dfn = iDfn + 1;
		// System.err.println("Opening file:\t" + dfn);
		openFileRead(vName + "/file" + dfn);
		readOligo(ol0); // ------------------------------------------- , "/file", fSize);
		while(strIn.available() > 0 || dfn < numberOfFiles) {
    		readOligo(ol1); // --------------------------------------- , "/file", fSize);
    		l[0] = l[1];
    		l[1] = ol1.match(ol0);
	        if (l[0] > l[1]) {
	        	u0 = l[0];
	        } else {
	        	u0 = l[1];
	        }
    		ol0.u0 = (byte) u0;
    		trg.writeOligo(ol0);
    		ol0 = new Oligo(ol1);
		}
		ol0.u0 = (byte) u0;
		trg.writeOligo(ol0);
		trg.close();
    }

    static public boolean delete(File path) {
        if(path.exists()) {
            File[] files = path.listFiles();
            for(int i=0; i<files.length; i++) {
               files[i].delete();
            }
        } else {
        	System.err.println("file not found:\t");
        }
        return(path.delete());
      }
    
    public void saveOligoP(int id, int chr, int start1, int strd1, int prbSize, int side1, int start2, int size2, byte[] mp, byte[] ref, byte[] prbs, int sidx) throws Exception {
    	// id2, chr, start1, maxScore[0]%2, size1, side1, start2, size2, mp, ref, prbs
    	boolean rc;
    	byte tmp;
    	float tmpf;
    	// int tmpq;
    	int countMiss = 0;
    	int[] sum = new int[prbSize];
    	float newScore = 1;
    	int start = 0;
    	int petSize = 0;
    	int end   = prbSize;
    	int ir = 0;
    	if (strd1 == 1) {
    		rc = true;
    	} else {
    		rc = false;
    	}
    	ir = 0;
    	while (ref[ir] < 0) {
    		ir++;
    	}
    	sum[ir] = prbs[ref[ir]];
    	if(sum[ir] < (-Parameters.seqSegPRB))
    		start = ir+1;
    	for (int i = ir + 1; i < prbSize; i++) {
    		if(ref[i] >= 0) {
        		sum[i] = sum[i-1] + prbs[i*4 + ref[i]]; 
        		if(sum[i] < (-Parameters.seqSegPRB)) {
        			if (prbs[i*4 + ref[i]] < (-Parameters.seqSegPRB))
        				start = i + 1;
        			else
        				start = i;
        			if (sum[i] < -(41 + Parameters.seqSegPRB))
        				sum[i] = -(41 + Parameters.seqSegPRB);
        		}
    		}
    	}
    	ir = prbSize - 1; 
    	while (ref[ir] < 0) {
    		ir--;
    	}
     	sum[ir] = prbs[(ir) * 4 + ref[ir]];
    	if (sum[ir] < (-Parameters.seqSegPRB))
    		end = ir-1;
    	for (int i = ir - 1; i >= 0; i--) {
    		if(ref[i] >= 0) {
    	   		sum[i] = sum[i+1] + prbs[i*4 + ref[i]]; 
        		if(sum[i] < (-Parameters.seqSegPRB)) {
        			if (prbs[i*4 + ref[i]] < (-Parameters.seqSegPRB))
        				end = i - 1;
        			else
        				end = i;
        			if (sum[i] < -(41 + Parameters.seqSegPRB))
        				sum[i] = -(41 + Parameters.seqSegPRB);
        		}
        	}
    	}
    	short size = (short) (end - start); 
    	int rf;
    	for (int i = start; i < end; i++) {
    		rf = ref[i];
    		if (rf >= 0) {
        		if (prbs[i*4 + rf] < (-Parameters.seqSegPRB))
        			countMiss++;
        		}
    	}
    	if (start1 < start2) {
    		petSize = Math.abs(start2 + size2 - start1);
    	} else {
    		petSize = Math.abs(start1 + prbSize - start2);
    	}
    	start1 += start;
    	if (size >= Parameters.seedSize && countMiss < 4) {
        	char[] st = new char[size];
        	for (int i = 0; i < prbSize; i++) {
        		rf = ref[i];
        		if (rf < 0)
        			rf = 4;
        		if(i >= start && i < end)
        			st[i-start] = ReferenceSequences.nucl[rf];
        	}
        	String s = new String(st); //.substring(start, end);
        	Oligo ol = new Oligo(s);
    		ol.prb   = new byte[4][size];
        	ol.size    = size;
        	ol.start   = start1;
        	ol.ch      = (byte) chr;
        	ol.id1     = id;
        	ol.setSide(side1);
       		ol.setRC(rc);
        	ol.setUM(true);
    		ol.prbSize = (short) size;
        	ol.u0 = mp[start];
        	if (petSize >= Parameters.petSpanMin && petSize <= Parameters.petSpanMax) {
        		ol.setPET(true);
            	ol.petSize = (short) petSize;
        	} else {
        		ol.setPET(false);
        	}
        	for (int i = 0; i < prbSize; i++) {
        		for (int b = 0; b < 4; b++) {
        			if (rc) {
    					if (i >= start && i < end) {
            				tmp = prbs[((end  + start - i - 1) * 4) + b];
                			ol.prb[b][i-start] = (byte) Consensus.getQCal(tmp, i);
    					}

        			} else {
    					if (i >= start && i < end) {
            				tmp = prbs[(i) * 4 + b];
                			ol.prb[b][i-start] = (byte) Consensus.getQCal(tmp, i);
    					}
        			}
        		}
        	}
        	countMiss = 0;
           	for (int i = start; i < start + Parameters.seedSize; i++) {
        		rf = ref[i];
        		if (rf < 0) {
        			tmp = -5;
        		} else {
               		if (rc) {
               			tmp = ol.prb[ref[i]][ol.size - (i-start) - 1];
               		} else {
               			tmp = ol.prb[ref[i]][i-start];
               		}
        		}
        		if (tmp > (-Parameters.q_Cal[i] + 2)) {
            		tmpf = Consensus.q2p(tmp);
            		newScore *= tmpf;
        		} else {
        			countMiss++;
        			if (countMiss > 3) {
        				break;
        			}
        		}
        	}
           	if (countMiss < 4) {
           		ol.weight = 1f - ((ol.u0 - ol.size) / 4 + countMiss) * 0.1f;
           		if (ol.isPET()) {
           			ol.weight += 0.2f;
           			if (ol.weight > 0.9)
           				ol.weight = 0.9f;
           			else if (ol.weight < 0.5f)
           				ol.weight = 0.5f;
           		} else {
           			if (ol.weight > 0.8)
           				ol.weight = 0.8f;
           			else if (ol.weight < 0.3f)
           				ol.weight = 0.3f;
           		}
           		sort_WriteOligoP(ol, sidx);
           	}
    	}
    	/*
    	System.err.print(id + "\t" + location + "\t" + strd + "\t" + prbSize + "\t" + start + "\t" + end + "\t" + newScore / size + "\t" + s + "\t");
    	for (int i = 0; i < prbSize; i++) {
        	System.err.print(ReferenceSequences.nucl[ref[i]]);
    	}
    	System.err.println();
    	for (int i = 0; i < 4 * prbSize; i++) {
        	System.err.print(prbs[i] + "\t");
    	}    	
    	System.err.println();
    	//*/
    }
}
