// package JavaCode;

import java.util.ArrayList;

/*
SLIDER: Local Short Reads DNA Sequence Alignment
Beta Release; Ver 0.1, 22th Mar. 2008
By: Nawar Malhis
Canada's Michael Smith Genome Sciences Centre
*/

public class Seq {
	
    static final char[] nucl  = {'A', 'C', 'G', 'T'}; // used to convert int to char
	static final int       A = 0;          // 
	static final int       C = 1;          // 
	static final int       G = 2;          // 
	static final int       T = 3;          // 
	static       float[] qCal = new float[81];

	public int        size;
	public int        prbSize;
	public int        l;
	public float[][]  p;
	public char[]     ps;
	public byte[][][] _q;
	public String     posibleSeq;
	public float      psp;
	public int[][]    alt;
	public int[]      used;  // the list of not good reads
	public byte[]     acc;
	public byte[]     count; // this value is the number of possible nucletids
	public int[]      mx;
	public int        totalUd;
	public float      sumw;
	public int        id;
	public int        side;
	public float[]    infoc;
	
	private static Oligo oligo;

	public ArrayList<Oligo> bunk; 
/*
	public Seq(int sz) {
		prbSize = (short) sz;
		size  = Parameters.seedSize;
		if(prbSize < Parameters.seedSize) {
			prbSize = (short) Parameters.seedSize;
		}
		p     = new float[prbSize][4];
		q     = new byte [2][prbSize][4];
		count = new byte [prbSize];
		acc   = new byte [prbSize];
		alt   = new int  [prbSize][4];
		used  = new int  [prbSize];
		mx    = new int  [prbSize];
		
		bunk        = new ArrayList<Oligo>(); 
	}
	//*/
	public Seq() {
		size  = Parameters.seedSize;
		prbSize = (short) Parameters.prbMxSize;
		p     = new float[prbSize][4];
		ps    = new char[prbSize];
		infoc = new float[prbSize];
		_q     = new byte [2][prbSize][4];
		count = new byte [prbSize];
		acc   = new byte [prbSize];
		alt   = new int  [prbSize][4];
		used  = new int  [prbSize];
		mx    = new int  [prbSize];
		
		bunk        = new ArrayList<Oligo>(); 
	}

	public float getP_MPS(int[][] histo, float[][] histoW, float[][] cnt, int idxOne) { 
		double rt = 1;
		for(int i = 0; i < size; i++) {
			mx[i] = 0;
			for (int j = 1; j < 4; j++) {
				if (p[i][mx[i]] < p[i][j])
					mx[i] = j;
			}
			rt = rt * p[i][mx[i]];
			if (p[i][mx[i]] < Parameters.seqSegNucP) {
				histoW[idxOne][i] += p[i][mx[i]];
				histo[idxOne][i]++;
				cnt[idxOne][i] += count[i];
			}
		}
		psp = (float) rt;
		return (float) rt;
	}
	
	public float getMPS()  throws Exception {
    	char[] mps = posibleSeq.toCharArray();
		float rt = 1;
		bunk.clear();
		for(int i = 0; i < size; i++) {
			mx[i] = 0;
			for (int j = 1; j < 4; j++) {
				if (p[i][mx[i]] < p[i][j])
					mx[i] = j;
			}
			mps[i] = nucl[mx[i]];
			rt = rt * p[i][mx[i]];
		}
		posibleSeq = new String(mps);
		oligo = new Oligo(posibleSeq);
		oligo.weight = rt;
		bunk.add(new Oligo(oligo));
		return rt;
	}
	
	public int solvePrbLine(int ll) throws Exception {
		size = Parameters.seedSize;
		char[] ps = new char[size];
		int x;
		l = ll;
		totalUd = 0;
		for (int i = 0; i < size; i++) {
			ps[i] = 0;
			count[i] = 0;
			used[i]  = -1;
			for (int j = 0; j < 4; j++) {
				alt[i][j] = -1;
				x = _q[l][i][j]; //
				p[i][j] = qCal[x + 40]; // (float) (1/(1+Math.pow(10,(((float)-x)/10))));
				// if (p[i][j] > (1-Parameters.seqSegNucP)) {
				if (_q[l][i][j] > (0-Parameters.seqSegPRB)) {
					if(ps[i] == 0) {
						ps[i] = nucl[j];
					}
					// if (p[i][j] < Parameters.seqSegNucP) {
					if (_q[l][i][j] < Parameters.seqSegPRB) {
						used[totalUd]  = i;
					}
					alt[i][count[i]] = j;
					count[i]++;
				}
			}
			if (count[i] > 0)
				ps[i] = nucl[alt[i][count[i]-1]];
			if (used[totalUd] > -1) {
				totalUd++;
			}
		}
		posibleSeq = new String(ps).substring(0, size); 
		int ret = 4;
		if (totalUd == 0) {
			ret = 0;
		} else if (totalUd <= Parameters.maxUDN) {
			ret = 1;
		} 
		return ret;
	}
	
	public int solvePrbLineDB(int ll) throws Exception {
		// size = Parameters.seedSize;
		// char[] ps = new char[size];
		l = ll;
		totalUd = 0;
		for (int i = 0; i < size; i++) {
			ps[i] = 0;
			count[i] = 0;
			used[i]  = -1;
			for (int j = 0; j < 4; j++) {
				alt[i][j] = -1;
				if (p[i][j] > (1-Parameters.seqSegNucP)) {
					if(ps[i] == 0) {
						ps[i] = nucl[j];
					}
					if (p[i][j] < Parameters.seqSegNucP) {
						used[totalUd]  = i;
					}
					alt[i][count[i]] = j;
					count[i]++;
				}
			}
			if (count[i] > 0)
				ps[i] = nucl[alt[i][count[i]-1]];
			if (used[totalUd] > -1) {
				totalUd++;
			}
		}
		posibleSeq = new String(ps).substring(0, prbSize); 
		if (totalUd == 0) {
			return 0;
		} else if (totalUd <= Parameters.maxUDN) {
			return 1;
		} 
		return -1;
	}

	public void infoContent(int ll) {
		for (int i = 0; i < size; i++) {
				infoc[i] = baseInfo(i);
		}
	}
	
	float baseInfo(int ix) {
		float info = 2;
		for (int j = 0; j < 4; j++) {
			if (p[ix][j] > 0)
				info +=  p[ix][j] * (Math.log(p[ix][j])/0.693147181);
		}
		if (info < 0)
			info = 0;
		return info;
	}
	
	public int getPSLQ() throws Exception {
		
		float pr;
   		int i = 0;
   		int i2;
   		int ib;
   		int base;
   		char[] olc;
   		pr = getMPS();
    	olc = posibleSeq.toCharArray();
   		i = 0;
   		while(i < size && used[i] >= 0) {
   			i2 = used[i];
   			base = ReferenceSequences.bp2Int(olc[used[i]]);
   			if (base != mx[i2]) {
   				System.err.println("\nMax != base\t" + i + "\t" + i2 + "\tmx\t" + mx[i2] + "\tbase\t" + base);
   			}
   			ib = 0;
   			while(ib < 4 && alt[i2][ib] >= 0) {
   				if (alt[i2][ib] != base) {
   					olc[i2] = nucl[alt[i2][ib]];
   		    		oligo = new Oligo(new String(olc));
   		    		psp = pr * p[i2][alt[i2][ib]] / p[i2][base];
   		    		if (psp > Parameters.minWeight * 10) {
   		    			oligo.weight = psp;
   		        		bunk.add(new Oligo(oligo));
   		    		}
   					// ------------------------------------------------
   					olc[i2] = nucl[base];
   				}
   				ib++;
   			}
   			i++;
   		}
    	return -1;
	}

	public int getPS() throws Exception {
		int t1;
		int xx = 1;
		sumw = 0;
    	char[] olc = posibleSeq.toCharArray();
    	int[] olcAlt = new int[posibleSeq.length()];
   		for (int i = 0; i < size; i++) {
    		olcAlt[i] = count[i] - 1;
    	}
   		bunk.clear();
    	while(xx == 1) {
    		xx = 0;
    		oligo = new Oligo(posibleSeq);
    		psp = 1;
    		for (int i = 0; i < size; i++) {
    			psp = psp * p[i][alt[i][olcAlt[i]]];
    		}
    		oligo.weight = psp;
    		sumw += psp;
    		if (psp > Parameters.minWeight) {
        		bunk.add(new Oligo(oligo));
       		}
    		for (int i = totalUd - 1; i >= 0; i--) {
    			if (olcAlt[used[i]] == 0) {
    				olcAlt[used[i]] = count[used[i]] - 1;
    				t1 = olcAlt[used[i]];
    				olc[used[i]] = nucl[alt[used[i]][t1]];
    			} else {
    				olcAlt[used[i]]--;
    				t1 = olcAlt[used[i]];
    				olc[used[i]] = nucl[alt[used[i]][t1]];
    				posibleSeq = new String(olc);
    				xx = 1;
    				break;
    			}
    		}
    	}
    	return -1;
	}

	public boolean isCrisp(int i, int bpat) {
		if (p[i][bpat] > Parameters.seqSegNucP)
			return true;
		return false;
	}
	public void display() throws Exception {
		System.out.println(posibleSeq + "\t" + trim(psp));
		for (int j = 0; j < 4; j++) {
			for (int i = 0; i < size; i++) {
				System.out.print(trim(p[i][j]) + "\t");
			}
			System.out.println();
		}
		System.out.println("--------------------------");
		for (int i = 0; i < size; i++) {
			System.out.print(i + "\t");
		}
		System.out.println();
		System.out.println("Used:\t" + totalUd);
		for (int i = 0; i < size; i++) {
			System.out.print(used[i] + "\t");
		}
		System.out.println();
		System.out.println("Count:");
		for (int i = 0; i < size; i++) {
			System.out.print(count[i] + "\t");
		}
		System.out.println();
		System.out.println("Alt:");
		for (int j = 0; j < 4; j++) {
			for (int i = 0; i < size; i++) {
				System.out.print(alt[i][j] + "\t");
			}
			System.out.println();
		}
		System.err.println((new Oligo(posibleSeq))); //.display();
	}
	
	private float trim(float f) {
		return ((float) ((int) (f * 1000))) / (float) 1000.0; 
	}

	public int  getBpInt(char b) {
    	if (b == 'A')
    	    return A;
    	if (b == 'C')
    	    return C;
    	if (b == 'G')
    	    return G;
    	if (b == 'T')
    	    return T;
    	return -1;
    }
	
	public char getMPBP(int prbPoint) {
		int most = 0;
		if (true) { //map.cnt[dr][l] > 0) {
			for(int i = 1; i < 4; i++) {
				if (p[prbPoint][i] > p[prbPoint][most])
					most = i;
			}
			if (p[prbPoint][most] > 0.8)
				return Character.toUpperCase(nucl[most]);
			return Character.toLowerCase(nucl[most]);
		} else {
			return '_';
		}
	}
	//
	
}
